{
    vm - A virtual machine for vmAsm assembly
    
    Based on the XVM Virtual Machine by Alex Varanese (ISBN 1-931841-57-8)     
    
    ---------------------------------------
    
    Copyright (c) 2010 Ben Golightly, based on code by Alex Varanese
    Copyright (c) 2002 Alex Varanese

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.
    
    ---------------------------------------
    
    The imaginatively named "vm" Virtual Machine is a typeless, procedural VM,
    with native support for functions. Use the "vmAsm" assembler to assemble
    programs into a vmx (VM eXecutable) binary file that vm can run.  
    
    ---------------------------------------

}
Unit
    Uses
        vmAsm
        
        // Until a host can register HAPIs through function pointers, you may
        // want to add other units to "uses" above and define your functions in
        // Vm_Init and Vm_CallHapiFunc

Const
    VM_VERSION = 20100629
    VM_STACK_SIZE = 1024 * 16
    
    CMP_E  =  0
    CMP_G  =  1
    CMP_L  = -1
    CMP_UNDEFINED = 9
    
    // Local HAPI funcions
    HAPI_MESSAGEBOX = 0
    HAPI_CONFIRM    = 1
    HAPI_FUNC0      = 2
    HAPI_FUNC1      = 3
    HAPI_FUNC2      = 4
    HAPI_FUNC3      = 5

Type vmAssembledStrings = Record
    sString : String
EndType ; export

Type vmStrings = Record
    sString : String = ""
EndType ; export

Type vmValues = Record
    opType : Integer = 0
    iValue : Integer = 0
    rValue : Real = 0.0
    sValue: ^vmStrings
    iOffsetIndex: Integer = 0
EndType ; export

Type vmInstrs = Record
    opCode   : Integer
    opCount  : Integer
    opList   : List of vmValues // Operand list - optimisation - change to array?
EndType ; export

Type vmFuncs = Record
    entryPoint : Integer
    paramCnt   : Integer
    dataSize   : Integer
    frameSize  : Integer
    name       : ^vmStrings
EndType ; export

Type vmRegisteredHapiFuncs = Record
    name: String
    id: Integer = -1
EndType ; export

Type vmHapiFuncs = Record
    name: String
    registeredFunc: ^vmRegisteredHapiFuncs
EndType ; export

Type vmScripts = Record
    // Main
    globalDataSize: Integer
    
    // Instructions Stream
    instrStream: List of vmInstrs
    instrStreamSize: Integer
    instrStreamCurrInstr: ^vmInstrs
    instrStreamCurrInstrId: Integer
    
    // Runtime Stack
    runStackElementList: Array [VM_STACK_SIZE] of ^vmValues
    runStackSize       : Integer
    runStackTopIndex   : Integer
    runStackFrameIndex : Integer
    runStackDepth      : Integer = 0
    
    // Important Globals
    runStackGlobalRet : ^vmValues
    runStackGlobalCmp : ^vmValues
    
    // Tables
    funcsList: List of vmFuncs
    hapiFuncsList : List of vmHapiFuncs
    stringsList: List of vmStrings
EndType ; export

Type vmResultStrings = Record
    s : String = ""
EndType ; export

Var
    vmResultStringList : List of vmResultStrings ; export
    
    vmScriptsList: List of vmScripts
    vmActiveScript: ^vmScripts
    vmRegisteredHapiFuncList: List of vmRegisteredHapiFuncs 
    
    vmErrors: Integer = 0 ; export
    
    vmDoneInit: Boolean = FALSE

Procedure Vm_Init()
Begin
    // Register your HAPI Functions here
    
    If vmDoneInit = TRUE then exit
    vmDoneInit = TRUE

    Vm_RegisterHapiFunc("MessageBox", HAPI_MESSAGEBOX)
    Vm_RegisterHapiFunc("Confirm", HAPI_CONFIRM)
    Vm_RegisterHapiFunc("Func0", HAPI_FUNC0)
    Vm_RegisterHapiFunc("Func1", HAPI_FUNC1)
    Vm_RegisterHapiFunc("Func2", HAPI_FUNC2)
    Vm_RegisterHapiFunc("Func3", HAPI_FUNC3)
End

Procedure Vm_CallHapiFunc(hfunc: ^vmRegisteredHapiFuncs)
Var
    tmpValue: ^vmValues
    iTmp: Integer
Begin
    // Define the behaviour of your HAPI Functions here
    //
    // Use Vm_Pop() to get function parameters
    //
    // For a usable data type, use:
    //   Vm_CoerceValueToString(value)
    //   Vm_CoerceValueToInteger(value)
    //   Vm_CoerceValueToFloat(value)
    //
    // To return a value, use:
    //   Vm_ReturnString(string)
    //   Vm_ReturnInteger(integer)
    //   Vm_ReturnFloat(real)
    
    Case hfunc.id of
        (HAPI_MESSAGEBOX):
            tmpValue = Vm_Pop()
            MessageBox(Vm_CoerceValueToString(tmpValue), "VM Message")
        
        (HAPI_CONFIRM):
            tmpValue = Vm_Pop()
            iTmp = MessageBox(Vm_CoerceValueToString(tmpValue), "VM Confirm?", MB_YESNO)
            Vm_ReturnInteger(iTmp)
            
        (HAPI_FUNC0):
            MessageBox("HAPI function 'Func 0' was called")
            
        (HAPI_FUNC1):
            MessageBox("HAPI function 'Func 1' was called")
            
        (HAPI_FUNC2):
            MessageBox("HAPI function 'Func 2' was called")
            
        (HAPI_FUNC3):
            MessageBox("HAPI function 'Func 3' was called")
                            
        Default:
            Vm_AddRuntimeError("Unregistered HAPI Function (code "+hfunc.id+")")
    EndCase
    
End

Procedure Vm_RegisterHapiFunc(funcName: String, funcId: Integer)
Var
    hfunc: ^vmRegisteredHapiFuncs
Begin
    funcName = Upper(funcName)
    
    hfunc = NewItem(vmRegisteredHapiFuncList)
    hfunc.name = funcName
    hfunc.id = funcId
     
End

Procedure Vm_ReturnString(s: String)
Begin
    vmActiveScript.runStackGlobalRet.opType = OP_TYPE_STRING_INDEX
    vmActiveScript.runStackGlobalRet.sValue.sString = s
End

Procedure Vm_ReturnInteger(i: Integer)
Begin
    vmActiveScript.runStackGlobalRet.opType = OP_TYPE_INT
    vmActiveScript.runStackGlobalRet.iValue = i
End

Procedure Vm_ReturnFloat(f: Real)
Begin
    vmActiveScript.runStackGlobalRet.opType = OP_TYPE_FLOAT
    vmActiveScript.runStackGlobalRet.rValue = f
End

Function Vm_GetOpType(index: Integer) : Integer
Var
    vmValue: ^vmValues
Begin
    vmValue = IndexedItem(vmActiveScript.instrStreamCurrInstr.opList, index)
    result = vmValue.opType
End

Function Vm_ResolveOpType(index: Integer) : Integer
Var
    vmValue: ^vmValues
Begin
    vmValue = Vm_ResolveOpValue(index)
    result = vmValue.opType
End

Function Vm_ResolveOpPointer(index: Integer) : ^vmValues
Var
    i: Integer
    vmValue: ^vmValues
Begin
    vmValue = IndexedItem(vmActiveScript.instrStreamCurrInstr.opList, index)

    If (vmValue.opType = OP_TYPE_REL_STACK_INDEX) or (vmValue.opType = OP_TYPE_ABS_STACK_INDEX) then
        i = Vm_ResolveStackIndex(index)
        result = vmActiveScript.runStackElementList[i]
    Else
        result = vmValue
    Endif
End

Function Vm_ResolveStackIndex(index: Integer) : Integer
Begin
    If index < 0 then result = index + vmActiveScript.runStackFrameIndex else result = index
End

Function Vm_GetOperand(index: Integer) : ^vmValues
Begin
    result = IndexedItem(vmActiveScript.instrStreamCurrInstr.opList, index)
End

Function Vm_GetStackValue(index: Integer) : ^vmValues
Begin
    result = vmActiveScript.runStackElementList[Vm_ResolveStackIndex(index)]
End

Procedure Vm_SetStackValue(index: Integer, setValue: ^vmValues)
Var
    vmValue: ^vmValues
Begin
    vmValue = vmActiveScript.runStackElementList[Vm_ResolveStackIndex(index)]
    vmValue.opType = setValue.opType
    vmValue.iValue = setValue.iValue
    vmValue.rValue = setValue.rValue
    If setValue.opType = OP_TYPE_STRING_INDEX then
        vmValue.sValue.sString = setValue.sValue.sString
    Endif
    vmValue.iOffsetIndex = setValue.iOffsetIndex 
End

Function Vm_ResolveOpStackIndex(index: Integer) : Integer
Var
    vmValue: ^vmValues
    stackValue: ^vmValues
    base, offset: Integer
Begin
    vmValue = IndexedItem(vmActiveScript.instrStreamCurrInstr.opList, index)
    
    Case vmValue.opType of
        (OP_TYPE_ABS_STACK_INDEX):
            result = vmValue.iValue 
        
        (OP_TYPE_REL_STACK_INDEX):
            base = vmValue.iValue
            offset = vmValue.iOffsetIndex
            stackValue = Vm_GetStackValue(offset)
            result = base - stackValue.iValue // +
        
        Default:
            result = 0
            MessageBox("ResolveOpStackIndex - case shouldn't happen")
             
            
    
    EndCase
End

Function Vm_ResolveOpValue(index: Integer) : ^vmValues
Var
    vmValue: ^vmValues
    i: Integer
Begin
    vmValue = IndexedItem(vmActiveScript.instrStreamCurrInstr.opList, index)
    
    Case vmValue.opType of
        (OP_TYPE_ABS_STACK_INDEX):
            i = Vm_ResolveOpStackIndex(index)
            result = Vm_GetStackValue(i)
        
        (OP_TYPE_REL_STACK_INDEX):
            i = Vm_ResolveOpStackIndex(index)
            result = Vm_GetStackValue(i)
    
        Default:
            result = vmValue
    EndCase
End

Function Vm_ResolveOpAsInt(index: Integer) : Integer
Var
    vmValue: ^vmValues
Begin
    vmValue = Vm_ResolveOpValue(index)
    result = Vm_CoerceValueToInt(vmValue)
End

Function Vm_ResolveOpAsFloat(index: Integer) : Real
Var
    vmValue: ^vmValues
Begin
    vmValue = Vm_ResolveOpValue(index)
    result = Vm_CoerceValueToFloat(vmValue)
End

Function Vm_ResolveOpAsString(index: Integer) : String
Var
    vmValue: ^vmValues
Begin
    vmValue = Vm_ResolveOpValue(index)
    result = Vm_CoerceValueToString(vmValue)
End

Function Vm_CoerceValueToInt(vmValue: ^vmValues) : Integer
Begin
    Case vmValue.opType of
        (OP_TYPE_INT): result = vmValue.iValue
        (OP_TYPE_FLOAT): result = ToInt(vmValue.rValue)
        (OP_TYPE_STRING_INDEX):
            If Asm_StringIsFloat(vmValue.sValue.sString) then
                result = ToInt(vmValue.sValue.sString)
            Else
                result = 0
            Endif
        default:
            result = 0
    EndCase
End

Function Vm_CoerceValueToFloat(vmValue: ^vmValues) : Real
Begin
    Case vmValue.opType of
        (OP_TYPE_INT): result = ToReal(vmValue.iValue)
        (OP_TYPE_FLOAT): result = vmValue.rValue
        (OP_TYPE_STRING_INDEX):
            If Asm_StringIsFloat(vmValue.sValue.sString) then
                result = ToReal(vmValue.sValue.sString)
            Else
                result = 0
            Endif
        default:
            result = 0.0
    EndCase
End

Function Vm_CoerceValueToString(vmValue: ^vmValues) : String
Begin
    Case vmValue.opType of
        (OP_TYPE_INT): result = ToString(vmValue.iValue)
        (OP_TYPE_FLOAT): result = ToString(vmValue.rValue)
        (OP_TYPE_STRING_INDEX): result = vmValue.sValue.sString
        default:
            result = ""
    EndCase
End

Procedure Vm_Push(vmValue: ^vmValues)
Var
    stackValue: ^vmValues
Begin
    stackValue = vmActiveScript.runStackElementList[vmActiveScript.runStackTopIndex]
    stackValue.opType = vmValue.opType
    stackValue.iValue = vmValue.iValue
    stackValue.rValue = vmValue.rValue
    If vmValue.opType = OP_TYPE_STRING_INDEX then
        stackValue.sValue.sString = vmValue.sValue.sString
    Endif
    stackValue.iOffsetIndex = vmValue.iOffsetIndex 
    
    Inc(vmActiveScript.runStackTopIndex)
End

Function Vm_Pop() : ^vmValues
Begin
    Dec(vmActiveScript.runStackTopIndex)
    result = vmActiveScript.runStackElementList[vmActiveScript.runStackTopIndex]
End

Procedure Vm_PushFrame(size: Integer)
Begin
    vmActiveScript.runStackTopIndex = vmActiveScript.runStackTopIndex + size
    vmActiveScript.runStackFrameIndex = vmActiveScript.runStackTopIndex 
End

Procedure Vm_PopFrame(size: Integer)
Begin
    vmActiveScript.runStackTopIndex = vmActiveScript.runStackTopIndex - size
End

Function Vm_GetFunc(index: Integer): ^vmFuncs
Begin
    result = IndexedItem(vmActiveScript.funcsList, index)
End

Function Vm_GetHapiFunc(index: Integer): ^vmHapiFuncs
Var
    vmHapiFunc: ^vmHapiFuncs
Begin
    Loop vmHapiFunc through vmActiveScript.hapiFuncsList
        If index = vmHapiFunc.registeredFunc.id then
            result = vmHapiFunc
            exit
        Endif
    EndLoop
    
    MessageBox("Vm_GetHapiFunc: Not found")
End

Procedure Vm_CopyValue(dest: ^vmValues, src: ^vmValues)
Begin
    If dest.opType = OP_TYPE_STRING_INDEX then
        //Free(dest.sValue)
        dest.sValue.sString = ""
    Endif

    dest.opType = src.opType
    dest.iValue = src.iValue
    dest.rValue = src.rValue
    dest.iOffsetIndex = src.iOffsetIndex
    
    If src.opType = OP_TYPE_STRING_INDEX then
        //New(dest.sValue)
        dest.sValue.sString = src.sValue.sString
    Endif
End

Function Vm_Load(mfi: Element) : ^vmScripts ; export
Var
    vmScript: ^vmScripts
    vmFunc: ^vmFuncs
    vmHapiFunc: ^vmHapiFuncs
    vmInstr: ^vmInstrs
    vmValue: ^vmValues
    vmString: ^vmStrings
    
    vmRegisteredHapiFunc: ^vmRegisteredHapiFuncs
    
    vmAssembledString: ^vmAssembledStrings
    vmAssembledStringList: List of vmAssembledStrings
    
    exe_id: String
    version: Integer
    globalSize: Integer
    
    sizeOfStringBlock: Integer
    sizeOfFunctionBlock: Integer
    sizeOfHapiBlock: Integer
    sizeOfInstrsBlock: Integer
    totalStrings: Integer
    totalFuncs: Integer
    totalHapiFuncs: Integer
    totalInstrs: Integer
    
    runStackGlobalRet: Integer
    runStackGlobalCmp: Integer
    
    m, i, j, c, bytes: Integer
    s: String
Begin
    // Init
    If vmDoneInit = FALSE then Vm_Init()
    
    vm_AddMessage("Loading vmx... (vm v"+VM_VERSION+")")
    m = Millisecs    

    // Check file is not NULL
    If (mfi = NULL) then
        vm_AddError("Input is NULL")
        exit
    Endif

    // Check 4-byte ID (vmxb)
    exe_id = ReadBStr(mfi, 4)
    If exe_id <> ASM_EXE_ID then
        vm_AddError("Invalid header ID")
        exit
    Endif
    exe_id = ""

    // Check version
    version = ReadInt(mfi)
    If version <> VM_VERSION then
        vm_AddError("Invalid vmx version")
        exit
    Endif
    
    // Stack size
    globalSize = ReadInt(mfi)
    
    // Position of _RET and _CMP on the stack
    runStackGlobalRet = ReadInt(mfi)
    runStackGlobalCmp = ReadInt(mfi)
    
    // Create script item
    vmScript = NewItem(vmScriptsList)
    vmScript.globalDataSize = globalSize
    vmActiveScript = vmScript
    
    // Sizes
    sizeOfStringBlock = ReadInt(mfi)
    sizeOfFunctionBlock = ReadInt(mfi)
    sizeOfHapiBlock = ReadInt(mfi)
    sizeOfInstrsBlock = ReadInt(mfi)
    
    // Temporary String Table
    totalStrings = ReadInt(mfi)
        
    For i = 0 to (totalStrings-1)
        bytes = ReadInt(mfi)
        s = ReadBStr(mfi, bytes)
        vmAssembledString = NewItem(vmAssembledStringList)
        vmAssembledString.sString = s
    Next
    
    // Function Table
    totalFuncs = ReadInt(mfi)
    
    For i = 0 to (totalFuncs-1)
        vmFunc = NewItem(vmScript.funcsList)
        bytes = ReadByte(mfi)
        New(vmFunc.name)
        vmFunc.name.sString = ReadBStr(mfi, bytes) 
        vmFunc.entryPoint = ReadInt(mfi)
        vmFunc.paramCnt = ReadByte(mfi)
        vmFunc.dataSize = ReadInt(mfi)
        vmFunc.frameSize = vmFunc.paramCnt + 1 + vmFunc.dataSize  
    Next
    
    // HAPI Function Table
    totalHapiFuncs = ReadInt(mfi)
    
    For i = 0 to (totalHapiFuncs-1)
        bytes = ReadByte(mfi)
        s = ReadBStr(mfi, bytes)
        vmHapiFunc = NewItem(vmScript.hapiFuncsList)
        vmHapiFunc.name = s  
    Next
    
    // Instruction Stream
    totalInstrs = ReadInt(mfi)
    vmScript.instrStreamSize = totalInstrs
    
    For i = 0 to (totalInstrs-1)
        vmInstr = NewItem(vmScript.instrStream)
        vmInstr.opCode = ReadByte(mfi)
        vmInstr.opCount = ReadByte(mfi)
        
        For j = 0 to (vmInstr.opCount-1)
            vmValue = NewItem(vmInstr.opList)
            vmValue.opType = ReadByte(mfi)
            
            Case vmValue.opType of
                (OP_TYPE_INT):              vmValue.iValue = ReadInt(mfi)
                (OP_TYPE_FLOAT):            vmValue.rValue = ReadReal(mfi)
                (OP_TYPE_STRING_INDEX):     vmValue.iValue = ReadInt(mfi)
                (OP_TYPE_INSTR_INDEX):      vmValue.iValue = ReadInt(mfi)
                (OP_TYPE_ABS_STACK_INDEX):  vmValue.iValue = ReadInt(mfi)
                (OP_TYPE_REL_STACK_INDEX):  vmValue.iValue = ReadInt(mfi) ; vmValue.iOffsetIndex = ReadInt(mfi)
                (OP_TYPE_FUNC_INDEX):       vmValue.iValue = ReadInt(mfi)
                (OP_TYPE_HAPI_INDEX):       vmValue.iValue = ReadInt(mfi)
            EndCase 

        Next 
    Next
    
    // Give each string operand a unique string buffer
    Loop vmInstr through vmScript.instrStream
        c = vmInstr.opCount
        
        Loop vmValue through vmInstr.opList
            If vmValue.opType = OP_TYPE_STRING_INDEX then
                vmAssembledString = IndexedItem(vmAssembledStringList, vmValue.iValue)
                New(vmValue.sValue)
                vmValue.sValue.sString = vmAssembledString.sString
            Else
                New(vmValue.sValue)
            Endif
        EndLoop
    EndLoop
    
    // Free temporary string list
    Loop vmAssembledString through vmAssembledStringList
        Free(vmAssembledString)
    EndLoop

    // Init the stack
    For i = 0 to (VM_STACK_SIZE - 1)
        New(vmScript.runStackElementList[i])
    Next
    
    // Init an empty string buffer for each stack element in a new block of memory
    For i = 0 to (VM_STACK_SIZE - 1)
        New(vmScript.runStackElementList[i].sValue)
    Next
    
    // Find _CMP and _RET
    vmScript.runStackGlobalCmp = vmScript.runStackElementList[runStackGlobalCmp]
    vmScript.runStackGlobalRet = vmScript.runStackElementList[runStackGlobalRet]
    
    // Push space for the globals
    Vm_PushFrame(vmScript.globalDataSize)
    
    result = vmScript
    
    // Bind HAPI Functions
    Loop vmHapiFunc through vmScript.hapiFuncsList
        Loop vmRegisteredHapiFunc through vmRegisteredHapiFuncList
            If vmHapiFunc.name = vmRegisteredHapiFunc.name then
                vmHapiFunc.registeredFunc = vmRegisteredHapiFunc
            Endif 
        EndLoop
    EndLoop
    
    // Done!
    m = Millisecs - m
    vm_AddMessage("Took "+m+" millisecs.")    
    
    
End

Procedure Vm_FreeScript(vmScript: ^vmScripts) ; export
Var
    vmHapiFunc: ^vmHapiFuncs
    vmInstr: ^vmInstrs
    vmValue: ^vmValues
    vmFunc: ^vmFuncs
    i, c: Integer
Begin

    // Free Functions
    Loop vmFunc through vmScript.funcsList
        vmFunc.name.sString = ""
        Free(vmFunc.name)
        Free(vmFunc)
    EndLoop

    // Free HAPI Functions
    Loop vmHapiFunc through vmScript.hapiFuncsList
        Free(vmHapiFunc)
    EndLoop

    // Free Instructions strings
    Loop vmInstr through vmScript.instrStream
        c = vmInstr.opCount        
        Loop vmValue through vmInstr.opList
            vmValue.sValue.sString = ""
            Free(vmValue.sValue)
        EndLoop
    EndLoop   

    // Free Instructions
    Loop vmInstr through vmScript.instrStream
        c = vmInstr.opCount        
        Loop vmValue through vmInstr.opList
            Free(vmValue)
        EndLoop
        Free(vmInstr)
    EndLoop

    // Free Stack string buffers
     For i = 0 to (VM_STACK_SIZE - 1)
        vmScript.runStackElementList[i].sValue.sString = ""
        Free(vmScript.runStackElementList[i].sValue)
    Next
    
    // Free stack
    For i = 0 to (VM_STACK_SIZE - 1)
        Free(vmScript.runStackElementList[i])
    Next
    
    // Finally free the script structure
    Free(vmScript)

End

Function Vm_GetFunction(vmScript: ^vmScripts, func: String) : ^vmFuncs ; export
Var
    vmFunc: ^vmFuncs
Begin
    func = Upper(func)

    Loop vmFunc through vmScript.funcsList    
        If vmFunc.name.sString = func then
            result = vmFunc
            exit
        Endif
    EndLoop
    
    Vm_AddError("Function '"+func+"' Not Found!")
End

Function Vm_RunFunction(vmScript: ^vmScripts, vmFunc: ^vmFuncs) : ^vmValues ; export
Var
    tmpValue: ^vmValues
    tmpIndex: Integer
Begin

    vmActiveScript = vmScript
    tmpIndex = vmScript.runStackFrameIndex
                
    /// Mimic a call to ret
            
    // Return Address
    New(tmpValue)
    tmpValue.iValue = -12345
    Vm_Push(tmpValue)
    Free(tmpValue)
                
    // Stack Frame
    Vm_PushFrame(vmFunc.dataSize + 2)
                
    // Function Index
    New(tmpValue)
    tmpValue.iValue = ItemInList(vmScript.funcsList, vmFunc)
    tmpValue.iOffsetIndex = vmActiveScript.runStackFrameIndex - 1
    vm_SetStackValue(vmActiveScript.runStackTopIndex - 1, tmpValue)
    Free(tmpValue)
        
    // Make the jump
    vmScript.instrStreamCurrInstr = IndexedItem(vmScript.instrStream, vmFunc.entryPoint)
    vmScript.instrStreamCurrInstrId = vmFunc.entryPoint
    Vm_Run()
            
    // Restore Frame Index
    vmScript.runStackFrameIndex = tmpIndex
    
    // Return the _ret value
    result = vmScript.runStackGlobalRet
End

Procedure Vm_Run()
Var
    opCode: Integer
    dest: ^vmValues
    src: ^vmValues
    tmpValue: ^vmValues
    func: ^vmFuncs
    hfunc: ^vmHapiFuncs

    frameIndex: Integer
    iTmp: Integer
    fTmp, fTmp2: Real
    sTmp, sTmp2: String    
    done: Boolean = FALSE
    incStream: Boolean
    
    debugInstr: ^vmInstrs
    
    i: Integer // tmp
Begin    
    
    While done = FALSE
    
        incStream = FALSE
        opCode = vmActiveScript.instrStreamCurrInstr.opCode
        
    
        // DEBUG - STACK VISUALISATION!
        {
              CLS()
              PenColor(ToRGBA(255,255,255))
              Text(10, 10, "vmActiveScript.runStackFrameIndex = "+vmActiveScript.runStackFrameIndex)
              Text(10, 25, "vmActiveScript.runStackTopIndex   = "+vmActiveScript.runStackTopIndex)
              Text(10, 40, "vmActiveScript.runStackDepth   = "+vmActiveScript.runStackDepth)
              Text(10, 55, "vmActiveScript.instrStreamCurrInstrId   = "+vmActiveScript.instrStreamCurrInstrId) 
    
          For i = 0 to 45
              Case vmActiveScript.runStackElementList[i].opType of
                  (OP_TYPE_INT): PenColor(ToRGBA(0,200,0))
                  (OP_TYPE_FLOAT): PenColor(ToRGBA(0,0,200))
                  (OP_TYPE_STRING_INDEX): PenColor(ToRGBA(200,0,0))
              EndCase
        
              If i > vmActiveScript.runStackTopIndex then PenColor(ToRGBA(100,100,100))
        
              Case vmActiveScript.runStackElementList[i].opType of
                  (OP_TYPE_INT): Text(30, 720 - (i*15)-5,           "Int:    "+vmActiveScript.runStackElementList[i].iValue+"         "+vmActiveScript.runStackElementList[i].iOffsetIndex)
                  (OP_TYPE_FLOAT):  Text(30, 720 - (i*15)-5,        "Real:   "+vmActiveScript.runStackElementList[i].rValue)
                  (OP_TYPE_STRING_INDEX):  Text(30, 720 - (i*15)-5, "String: "+vmActiveScript.runStackElementList[i].sValue.sString)
              EndCase
          Next
          Flip
          Pause(10)
          }
    
    
                    
        Case opCode of
            (INSTR_MOV):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                //If Vm_ResolveOpPointer(0) <> Vm_ResolveOpPointer(1) then
                    Vm_CopyValue(dest, src)
                //Endif
                incStream = TRUE

            (INSTR_PUSH):
                src = Vm_ResolveOpValue(0)
                Vm_Push(src)
                incStream = TRUE

            (INSTR_POP):
                dest = Vm_ResolveOpValue(0)
                vm_CopyValue(dest, Vm_Pop())
                incStream = TRUE
            
            (INSTR_ADD):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                If dest.opType = OP_TYPE_INT then
                    dest.iValue = dest.iValue + Vm_CoerceValueToInt(src);
                Else
                    dest.rValue = Vm_CoerceValueToFloat(dest) + Vm_CoerceValueToFloat(src);
                    dest.opType = OP_TYPE_FLOAT
                Endif
                incStream = TRUE
            
            (INSTR_SUB):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                If dest.opType = OP_TYPE_INT then
                    dest.iValue = dest.iValue - Vm_CoerceValueToInt(src);
                Else
                    dest.rValue = Vm_CoerceValueToFloat(dest) - Vm_CoerceValueToFloat(src);
                    dest.opType = OP_TYPE_FLOAT
                Endif
                incStream = TRUE
                
            (INSTR_MUL):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                If dest.opType = OP_TYPE_INT then
                    dest.iValue = dest.iValue * Vm_CoerceValueToInt(src);
                Else
                    dest.rValue = Vm_CoerceValueToFloat(dest) * Vm_CoerceValueToFloat(src);
                    dest.opType = OP_TYPE_FLOAT
                Endif
                incStream = TRUE
                
            (INSTR_DIV):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                If dest.opType = OP_TYPE_INT then
                    iTmp = Vm_CoerceValueToInt(src)
                    If iTmp = 0 then
                        Vm_AddRuntimeError("Divide by zero")
                    Else
                        dest.iValue = dest.iValue / iTmp;
                    Endif
                Else
                    fTmp = Vm_CoerceValueToFloat(src)
                    If fTmp = 0.0 then
                        Vm_AddRuntimeError("Divide by zero")
                    Else
                        dest.rValue = Vm_CoerceValueToFloat(dest) / fTmp;
                        dest.opType = OP_TYPE_FLOAT
                    Endif
                Endif
                incStream = TRUE
            
            (INSTR_MOD):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                iTmp = Vm_CoerceValueToInt(src)
                If iTmp = 0 then
                    Vm_AddRuntimeError("Divide by zero")
                Else 
                    dest.iValue = Vm_CoerceValueToInt(dest) mod iTmp
                Endif 
                incStream = TRUE
            
            (INSTR_ABS):
                dest = Vm_ResolveOpValue(0)
                If dest.opType = OP_TYPE_INT then
                    dest.iValue = Abs(dest.iValue)
                Else
                    dest.rValue = Abs(Vm_CoerceValueToFloat(dest));
                    dest.opType = OP_TYPE_FLOAT
                Endif
                 incStream = TRUE
            
            (INSTR_EXP):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                If dest.opType = OP_TYPE_INT then
                    dest.iValue = (dest.iValue)^(Vm_CoerceValueToInt(src))
                Else
                    dest.rValue = (Vm_CoerceValueToFloat(dest))^(Vm_CoerceValueToFloat(src));
                    dest.opType = OP_TYPE_FLOAT
                Endif
                incStream = TRUE
            
            (INSTR_NEG):
                dest = Vm_ResolveOpValue(0)
                If dest.opType = OP_TYPE_INT then
                    dest.iValue = -(dest.iValue)
                Else
                    dest.rValue = -(Vm_CoerceValueToFloat(dest));
                    dest.opType = OP_TYPE_FLOAT
                Endif
                 incStream = TRUE
                 
            (INSTR_INC):
                dest = Vm_ResolveOpValue(0)
                If dest.opType = OP_TYPE_INT then
                    Inc(dest.iValue)
                Else
                    dest.rValue = Vm_CoerceValueToFloat(dest)+1.0;
                    dest.opType = OP_TYPE_FLOAT
                Endif
                incStream = TRUE
                
            (INSTR_DEC):
                dest = Vm_ResolveOpValue(0)
                If dest.opType = OP_TYPE_INT then
                    Dec(dest.iValue)
                Else
                    dest.rValue = Vm_CoerceValueToFloat(dest)-1.0;
                    dest.opType = OP_TYPE_FLOAT
                Endif
                incStream = TRUE
            
            (INSTR_AND):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                dest.iValue = Vm_CoerceValueToInt(dest) and Vm_CoerceValueToInt(src)
                dest.opType = OP_TYPE_INT
                incStream = TRUE
                
            (INSTR_OR):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                dest.iValue = Vm_CoerceValueToInt(dest) or Vm_CoerceValueToInt(src)
                dest.opType = OP_TYPE_INT
                incStream = TRUE
                
            (INSTR_XOR):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                dest.iValue = Vm_CoerceValueToInt(dest) xor Vm_CoerceValueToInt(src)
                dest.opType = OP_TYPE_INT
                incStream = TRUE
                
            (INSTR_NOT):
                dest = Vm_ResolveOpValue(0)
                dest.iValue = not Vm_CoerceValueToInt(dest)
                dest.opType = OP_TYPE_INT
                incStream = TRUE
                
            (INSTR_SHL):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                dest.iValue = Vm_CoerceValueToInt(dest) Shl Vm_CoerceValueToInt(src)
                dest.opType = OP_TYPE_INT
                incStream = TRUE
                
            (INSTR_SHR):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                dest.iValue = Vm_CoerceValueToInt(dest) Shr Vm_CoerceValueToInt(src)
                dest.opType = OP_TYPE_INT
                incStream = TRUE
            
            (INSTR_CALL):
            
                src  = Vm_ResolveOpValue(0)
                func = Vm_GetFunc(src.iValue)
                
                // Return Address
                New(tmpValue)
                tmpValue.iValue = vmActiveScript.instrStreamCurrInstrId + 1
                Vm_Push(tmpValue)
                Free(tmpValue)
                
                // Stack Frame
                Vm_PushFrame(func.dataSize + 1)
                
                // Function Index
                New(tmpValue)
                tmpValue.iValue = src.iValue
                tmpValue.iOffsetIndex = vmActiveScript.runStackFrameIndex - 1
                vm_SetStackValue(vmActiveScript.runStackTopIndex - 1, tmpValue)
                Free(tmpValue)
                
                // Make the jump
                vmActiveScript.instrStreamCurrInstr = IndexedItem(vmActiveScript.instrStream, func.entryPoint)
                vmActiveScript.instrStreamCurrInstrId = func.entryPoint
                Inc(vmActiveScript.runStackDepth)
                

            (INSTR_RET):       
                
                If vmActiveScript.runStackDepth = 0 then
                    // Function Index
                    tmpValue = Vm_Pop()
                    func = Vm_GetFunc(tmpValue.iValue)
                    Vm_Pop()
                    Vm_PopFrame(func.frameSize)
                
                    // Exit from top-level function
                    done = TRUE
                Else
                    // Function Index
                    tmpValue = Vm_Pop()
                    func = Vm_GetFunc(tmpValue.iValue)
                    frameIndex = tmpValue.iOffsetIndex - func.frameSize

                    // Return Address
                    tmpValue = vm_GetStackValue(vmActiveScript.runStackTopIndex - (func.dataSize + 1))

                    // Pop frame & return address
                    Vm_PopFrame(func.frameSize)
                
                    // Restore previous frame index
                    vmActiveScript.runStackFrameIndex = frameIndex
                
                    // Make the jump
                    vmActiveScript.instrStreamCurrInstr = IndexedItem(vmActiveScript.instrStream, tmpValue.iValue)
                    vmActiveScript.instrStreamCurrInstrId = tmpValue.iValue
                    Dec(vmActiveScript.runStackDepth)
                Endif
                
            
            (INSTR_HAPI):
                src = Vm_ResolveOpValue(0)
                hfunc = Vm_GetHapiFunc(src.iValue)
                Vm_CallHapiFunc(hfunc.registeredFunc)
                incStream = TRUE
            
            (INSTR_JMP):

                src = Vm_ResolveOpValue(0)
                vmActiveScript.instrStreamCurrInstr = IndexedItem(vmActiveScript.instrStream, src.iValue)
                vmActiveScript.instrStreamCurrInstrId = src.iValue
            
            (INSTR_CMP):
                dest = Vm_ResolveOpValue(0)
                src  = Vm_ResolveOpValue(1)
                
                //MessageBox(""+dest.iOffsetIndex+", "+ItemInList(src))
                
                If dest.opType = OP_TYPE_INT then
                    iTmp = Vm_CoerceValueToInt(src)
                    If dest.iValue = iTmp then
                        vmActiveScript.runStackGlobalCmp.iValue = CMP_E
                    Else
                        If dest.iValue < iTmp then
                            vmActiveScript.runStackGlobalCmp.iValue = CMP_L
                        Else
                            vmActiveScript.runStackGlobalCmp.iValue = CMP_G
                        Endif
                    Endif
                Endif
                If dest.opType = OP_TYPE_STRING_INDEX then
                    sTmp = Vm_CoerceValueToString(dest)
                    sTmp2 = Vm_CoerceValueToString(src)
                    If sTmp = sTmp2 then
                        vmActiveScript.runStackGlobalCmp.iValue = CMP_E
                    Else
                        vmActiveScript.runStackGlobalCmp.iValue = CMP_UNDEFINED
                    Endif
                Endif
                If dest.opType = OP_TYPE_FLOAT then
                    fTmp = Vm_CoerceValueToFloat(dest)
                    fTmp2 = Vm_CoerceValueToFloat(src)
                    If fTmp = fTmp2 then
                        vmActiveScript.runStackGlobalCmp.iValue = CMP_E
                    Else
                        If fTmp < fTmp2 then
                            vmActiveScript.runStackGlobalCmp.iValue = CMP_L
                        Else
                            vmActiveScript.runStackGlobalCmp.iValue = CMP_G
                        Endif
                    Endif
                Endif
                incStream = TRUE
            

            (INSTR_JE):
                If (vmActiveScript.runStackGlobalCmp.iValue = CMP_E) then
                    src = Vm_ResolveOpValue(0)
                    vmActiveScript.instrStreamCurrInstr = IndexedItem(vmActiveScript.instrStream, src.iValue)
                    vmActiveScript.instrStreamCurrInstrId = src.iValue
                Else
                    incStream = TRUE
                Endif

            (INSTR_JNE):
                If (vmActiveScript.runStackGlobalCmp.iValue <> CMP_E) then
                    src = Vm_ResolveOpValue(0)
                    vmActiveScript.instrStreamCurrInstr = IndexedItem(vmActiveScript.instrStream, src.iValue)
                    vmActiveScript.instrStreamCurrInstrId = src.iValue
                Else
                    incStream = TRUE
                Endif
                
            (INSTR_JG):
                If (vmActiveScript.runStackGlobalCmp.iValue = CMP_G) then
                    src = Vm_ResolveOpValue(0)
                    vmActiveScript.instrStreamCurrInstr = IndexedItem(vmActiveScript.instrStream, src.iValue)
                    vmActiveScript.instrStreamCurrInstrId = src.iValue
                Else
                    incStream = TRUE
                Endif
                
            (INSTR_JL):
                If (vmActiveScript.runStackGlobalCmp.iValue = CMP_L) then
                    src = Vm_ResolveOpValue(0)
                    vmActiveScript.instrStreamCurrInstr = IndexedItem(vmActiveScript.instrStream, src.iValue)
                    vmActiveScript.instrStreamCurrInstrId = src.iValue
                Else
                    incStream = TRUE
                Endif
                
            (INSTR_JGE):
                If (vmActiveScript.runStackGlobalCmp.iValue = CMP_E) or (vmActiveScript.runStackGlobalCmp.iValue = CMP_G) then
                    src = Vm_ResolveOpValue(0)
                    vmActiveScript.instrStreamCurrInstr = IndexedItem(vmActiveScript.instrStream, src.iValue)
                    vmActiveScript.instrStreamCurrInstrId = src.iValue
                Else
                    incStream = TRUE
                Endif

            (INSTR_JLE):
                If (vmActiveScript.runStackGlobalCmp.iValue = CMP_E) or (vmActiveScript.runStackGlobalCmp.iValue = CMP_L) then
                    src = Vm_ResolveOpValue(0)
                    vmActiveScript.instrStreamCurrInstr = IndexedItem(vmActiveScript.instrStream, src.iValue)
                    vmActiveScript.instrStreamCurrInstrId = src.iValue
                Else
                    incStream = TRUE
                Endif
            
            (INSTR_CONCAT):
                dest = Vm_ResolveOpValue(0)
                src = Vm_ResolveOpValue(1)
                dest.sValue.sString = Vm_CoerceValueToString(dest) + Vm_CoerceValueToString(src)
                incStream = TRUE

            (INSTR_TERM):
                done = TRUE
            
            (INSTR_FLOAT):
                dest = Vm_ResolveOpValue(0)
                dest.rValue = Vm_CoerceValueToFloat(dest)
                dest.opType = OP_TYPE_FLOAT
                incStream = TRUE

            (INSTR_INT):
                dest = Vm_ResolveOpValue(0)
                dest.rValue = Vm_CoerceValueToInt(dest)
                dest.opType = OP_TYPE_INT
                incStream = TRUE
                
            (INSTR_STRING):
                dest = Vm_ResolveOpValue(0)
                dest.rValue = Vm_CoerceValueToString(dest)
                dest.opType = OP_TYPE_STRING_INDEX
                incStream = TRUE        
            
            Default:
                Vm_AddRuntimeError("Unkown instruction")
        EndCase
        
        If vmActiveScript.instrStreamCurrInstrId = (vmActiveScript.instrStreamSize-1) then
            done = TRUE
        Else
            If (incStream) then
                Inc(vmActiveScript.instrStreamCurrInstrId)
                vmActiveScript.instrStreamCurrInstr = NextItem(vmActiveScript.instrStream, vmActiveScript.instrStreamCurrInstr)
            Endif
        Endif   
    Wend
    
End

Procedure Vm_AddError(s: String)
Begin
    Inc(vmErrors)
    Vm_AddMessage("Error: "+s)
End

Procedure Vm_AddRuntimeError(s: String)
Begin
    Inc(vmErrors)
    Vm_AddMessage("Runtime Error: "+s+" in script "+ToInt(ItemInList(vmScriptsList, vmActiveScript))+", offset "+vmActiveScript.instrStreamCurrInstrId)
End

Procedure Vm_AddMessage(s: String)
Var
    vmResultString: ^vmResultStrings
Begin
    vmResultString = NewItem(vmResultStringList)
    If Assigned(vmResultString) then
        vmResultString.s = s
    Else
        MessageBox("There was an error adding an error.")
    Endif
End

Procedure Vm_ClearMessages() ; export
Var
    vmResultString: ^vmResultStrings
Begin
    Loop vmResultString through vmResultStringList
        Free(vmResultString)
    EndLoop
    vmErrors = 0
End

Begin
End