{
    vmAsm - An assembler for the "vm" virtual machine
    
    Based on the XASM Assembler by Alex Varanese (ISBN 1-931841-57-8)     
    
    ---------------------------------------
    
    Copyright (c) 2010 Ben Golightly, based on code by Alex Varanese
    Copyright (c) 2002 Alex Varanese

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.
    
    ---------------------------------------

}
Unit

Const
    
    ASM_EXE_ID = "vmxb"    ; export // 4 bytes - Virtual Machine Executionable Binary
    ASM_VERSION = 20100629 ; export  
    
    // Operand Flags
    OP_FLAG_INT     = 1  // Integer Literal
    OP_FLAG_FLOAT   = 2  // Float Literal
    OP_FLAG_STRING  = 4  // String Literal
    OP_FLAG_MEMREF  = 8  // Variable/array index
    OP_FLAG_LABEL   = 16 // Line-label
    OP_FLAG_FUNC    = 32 // Function Name
    OP_FLAG_HAPI    = 64 // HAPI Function

    // Assembled OpCodes    
    OP_TYPE_INT               = 0 ; export  // Integer literal value
    OP_TYPE_FLOAT             = 1 ; export  // Floating-point literal value
    OP_TYPE_STRING_INDEX      = 2 ; export  // String index value
    OP_TYPE_ABS_STACK_INDEX   = 3 ; export  // Absolute array index
    OP_TYPE_REL_STACK_INDEX   = 4 ; export  // Relative array index
    OP_TYPE_INSTR_INDEX       = 5 ; export  // Instruction index
    OP_TYPE_FUNC_INDEX        = 6 ; export  // Function index
    OP_TYPE_HAPI_INDEX        = 7 ; export  // Host API call index
    
    // Instructions
    INSTR_MOV  = 0  ; export
    INSTR_PUSH = 1  ; export
    INSTR_POP  = 2  ; export   
    INSTR_ADD  = 3  ; export
    INSTR_SUB  = 4  ; export
    INSTR_MUL  = 5  ; export
    INSTR_DIV  = 6  ; export
    INSTR_MOD  = 7  ; export
    INSTR_ABS  = 8  ; export
    INSTR_EXP  = 9  ; export
    INSTR_NEG  = 10 ; export
    INSTR_INC  = 11 ; export
    INSTR_DEC  = 12 ; export   
    INSTR_AND  = 13 ; export
    INSTR_OR   = 14 ; export
    INSTR_XOR  = 15 ; export
    INSTR_NOT  = 16 ; export
    INSTR_SHL  = 17 ; export
    INSTR_SHR  = 18 ; export   
    INSTR_CALL = 19 ; export
    INSTR_RET  = 20 ; export
    INSTR_HAPI = 21 ; export   
    INSTR_JMP  = 22 ; export
    INSTR_CMP  = 24 ; export
    INSTR_JE   = 25 ; export
    INSTR_JNE  = 26 ; export
    INSTR_JG   = 27 ; export
    INSTR_JL   = 28 ; export
    INSTR_JGE  = 29 ; export
    INSTR_JLE  = 30 ; export
    INSTR_CONCAT = 31 ; export
    INSTR_TERM   = 32 ; export
    INSTR_FLOAT  = 33 ; export // coerces variables to float
    INSTR_INT    = 34 ; export // coerces variables to integer
    INSTR_STRING = 35 ; export // coerces variables to string
    MAX_INSTRS   = 36 // Max instructions in the lookup table
    
    // Tokens
    TOKEN_INVALID       = 0
    TOKEN_INT           = 1
    TOKEN_FLOAT         = 2
    TOKEN_STRING        = 3 // does not include quotes
    TOKEN_QUOTE         = 4 // "
    TOKEN_IDENT         = 5
    TOKEN_COLON         = 6
    TOKEN_OPEN_BRACKET  = 7 // [
    TOKEN_CLOSE_BRACKET = 8 // ]
    TOKEN_COMMA         = 9
    TOKEN_OPEN_BRACE    = 10 // {
    TOKEN_CLOSE_BRACE   = 11 // }
    TOKEN_NEWLINE       = 12
    TOKEN_INSTR         = 13 // Instruction
    TOKEN_VAR           = 14 // Var directive
    TOKEN_FUNC          = 15
    TOKEN_PARAM         = 16 // Param directive
    TOKEN_END_STREAM    = 17 // End of token stream
    
    // Lexer States
    LEXER_STATE_NO_STR  = 0
    LEXER_STATE_IN_STR  = 1
    LEXER_STATE_END_STR = 2
        

Type asmStrings = Record
    sString : String
EndType

Type asmOperands = Record
    opType : Integer = 0
    iValue : Integer = 0
    rValue : Real = 0.0
    iOffsetIndex: Integer = 0
EndType

Type asmInstrs = Record
    opCode   : Integer
    opCount  : Integer
    opList   : List of asmOperands
EndType

Type asmFuncs = Record
    index    : Integer
    name     : String
    entry    : Integer
    paramCnt : Integer
    dataSize : Integer
EndType

Type asmSymbols = Record
    index      : Integer
    ident      : String
    size       : Integer // 1 for variables, N for arrays
    isArray    : Boolean
    stackIndex : Integer
    funcIndex  : Integer
EndType

Type asmLabels = Record
    index       : Integer
    ident       : String
    targetIndex : Integer
    funcIndex   : Integer
EndType

Type asmInstrLookups = Record
    mnemonic : String = ""
    opCode   : Integer
    opCount  : Integer
    opList   : List of asmOperands
EndType

Type asmSourceLines = Record
    s : String
EndType

Type asmResultStrings = Record
    s : String = ""
EndType ; export

Var
    asmStringList   : List of asmStrings
    asmHapiFuncList : List of asmStrings
    asmFuncList     : List of asmFuncs
    asmSymbolList   : List of asmSymbols
    asmLabelList    : List of asmLabels
    asmInstrLookup  : Array [MAX_INSTRS] of ^asmInstrLookups
    asmSourceList   : List of asmSourceLines
    asmCurrLine     : ^asmSourceLines
    
    asmResultStringList : List of asmResultStrings ; export
    
    asmDoneInit: Boolean = FALSE 
    
    asmLexer_index0    : Integer = 0
    asmLexer_index1    : Integer = 0
    asmLexer_currToken : Integer
    asmLexer_currLexeme: String = ""
    asmLexer_currState : Integer
    
    asmParser_globalDataSize: Integer = 0
    asmParser_instrStreamSize: Integer = 0
    asmParser_currInstrIndex: Integer = 0
    asmParser_isFuncActive : Boolean
    asmParser_currFunc: ^asmFuncs
    asmParser_currInstr: ^asmInstrLookups
    asmParser_currFuncName : String
    asmParser_currFuncIndex : Integer
    asmParser_currFuncParamCount : Integer
    asmParser_currFuncLocalDataSize : Integer
    
    asmErrors: Integer = 0 ; export

Function Asm_Assemble(mfi: Element, info: Boolean = TRUE) : Element ; export
{
    Asm_Assemble
    mfi: MemFile containing assembly code to assemble
    Returns a MemFile containing assembled machine code
    
    * Assembler is init'd if it hasn't been already
    * Each Line of code is put into the asmSourceList list
}
Var
    m: Integer
    s: String
    i: Integer
    asmSourceLine: ^asmSourceLines    
    
    funcName: String
    funcEntry: Integer
    funcIndex: Integer
    instruction: Integer
    ident: String
    size, stackIndex, targetIndex, offsetIndex: Integer
    opTypes: Integer
    sIndex: Integer
    isArray: Boolean
    arrayIndex: Integer
    arrayIndexIdent: String
    
    instrStream: List of asmInstrs
    asmInstr: ^asmInstrs
    asmLookupOperand: ^asmOperands
    asmOperand: ^asmOperands
    asmSymbol, asmSymbol2: ^asmSymbols
    asmLabel: ^asmLabels
    asmFunc: ^asmFuncs
    asmString: ^asmStrings
    
    tmp: Element
    asmResultString: ^asmResultStrings
    
    infoVariables: Integer
    infoArrays: Integer
    infoGlobals: Integer
    
    bytes: Integer
    
    isMainDefined: Boolean
    
Begin
    m = Millisecs

    If asmDoneInit = FALSE then Asm_Init()
    
    // Clear result string
    Loop asmResultString through asmResultStringList
        Free(asmResultString)
    EndLoop
    
    Asm_AddMessage("Assembling... (assembler v"+ToString(ASM_VERSION)+")")
    
    If mfi = NULL then
        Asm_AddMessage("Fatal Error: No data")
        exit
    Endif
    
    result = CreateMemFile()
    If result = NULL then
        Asm_AddMessage("Fatal Error: No output buffer")
        exit
    Endif
    
    // Strip whitespace & comments
    While Not EOF(mfi)
        s = ReadLn(mfi)
        s = Asm_StripComments(s)
        s = Asm_StripWhitespace(s)        
        asmSourceLine = NewItem(asmSourceList)
        asmSourceLine.s = s + Chr(10) // newline
    Wend
    
    // Add return global
    asmSourceLine = NewItemAtStart(asmSourceList)
    asmSourceLine.s = "Var _ret"+Chr(10)
    
    // Add compare global
    asmSourceLine = NewItemAtStart(asmSourceList)
    asmSourceLine.s = "Var _cmp"+Chr(10)   
    
    // Resets
    Asm_ResetLexer()
    asmParser_globalDataSize = 0    
    asmParser_instrStreamSize = 0
    asmParser_isFuncActive = FALSE
    asmErrors = 0
    
    
    
    
    // Parse Assembly - Pass 1
    // -------------------------------------------------------------------------
    
    Asm_AddMessage("Pass 1")
    
    While (TRUE)
        If (Asm_GetNextToken() = TOKEN_END_STREAM) then break 

        Case asmLexer_currToken of    
            
            // TOKEN_FUNC (directive)
            (TOKEN_FUNC):
                If (asmParser_isFuncActive) then Asm_AddCodeError("Cannot nest functions"); break
                If (Asm_GetNextToken() <> TOKEN_IDENT) then Asm_AddExpectedError("identifier") ; break
                
                funcName = asmLexer_currLexeme
                funcEntry = asmParser_instrStreamSize
                funcIndex = asm_AddFunc(funcName, funcEntry)
                
                If (funcIndex < 0) then Asm_AddCodeError("Function '"+funcName+"' is already defined") ; break
                
                asmParser_isFuncActive = TRUE
                asmParser_currFuncName = funcName
                asmParser_currFuncIndex = funcIndex
                asmParser_currFuncParamCount = 0
                asmParser_currFuncLocalDataSize = 0
                
                While (Asm_GetNextToken() = TOKEN_NEWLINE)
                Wend
                
                If (asmLexer_currToken <> TOKEN_OPEN_BRACE) then Asm_AddExpectedError("{"); break

                // Funcs are automatically appended with a Ret instr
                Inc(asmParser_instrStreamSize)
                
                
            // TOKEN_OPEN_BRACE (directive)
            (TOKEN_OPEN_BRACE):
                Asm_AddUnexpectedError("{") ; break
            
            // TOKEN_CLOSE_BRACE (directive)
            (TOKEN_CLOSE_BRACE):
                If (asmParser_isFuncActive = FALSE) then Asm_AddUnexpectedError("}")
                
                asmParser_currFunc = IndexedItem(asmFuncList, asmParser_currFuncIndex)
                asmParser_currFunc.paramCnt = asmParser_currFuncParamCount
                asmParser_currFunc.dataSize = asmParser_currFuncLocalDataSize
                
                asmParser_isFuncActive = FALSE
                //Inc(asmParser_instrStreamSize)
            
            
            // TOKEN_VAR (directive)
            (TOKEN_VAR):
                If (Asm_GetNextToken() <> TOKEN_IDENT) then
                    Asm_AddExpectedError("identifier")
                    If (asmLexer_currToken = TOKEN_INSTR) Asm_AddMessage("("+asmLexer_currLexeme+" is a reserved instruction)")
                    If (asmLexer_currToken = TOKEN_FUNC) Asm_AddMessage("("+asmLexer_currLexeme+" is a reserved directive)")
                    If (asmLexer_currToken = TOKEN_PARAM) Asm_AddMessage("("+asmLexer_currLexeme+" is a reserved directive)")
                    break
                Endif
                
                ident = asmLexer_currLexeme
                size = 1
                isArray = FALSE
                
                If (Asm_LexerLookAhead() = '[') then                    
                    If (Asm_GetNextToken() <> TOKEN_OPEN_BRACKET) then Asm_AddExpectedError("["); break
                    If (Asm_GetNextToken() <> TOKEN_INT) then Asm_AddCodeError("Array sizes must be integer and constant"); break
                    
                    isArray = TRUE
                    size = ToInt(asmLexer_currLexeme)
                    If size <= 0 then Asm_AddCodeError("Array sizes must be non-zero and positive"); break
                    If (Asm_GetNextToken() <> TOKEN_CLOSE_BRACKET) then Asm_AddExpectedError("]"); break
                Endif
                
                If (asmParser_isFuncActive) then
                    stackIndex = -(asmParser_currFuncLocalDataSize + 2)
                Else
                    stackIndex = asmParser_globalDataSize
                Endif
                
                If (Asm_AddSymbol(ident, size, stackIndex, asmParser_currFuncIndex, isArray) < 0) then
                    Asm_AddCodeError("Variable "+ident+" is already defined") ; break
                Endif
                
                If (asmParser_isFuncActive) then
                    asmParser_currFuncLocalDataSize = asmParser_currFuncLocalDataSize + size
                Else
                    asmParser_globalDataSize = asmParser_globalDataSize + size
                Endif
                
                If (Asm_GetNextToken() <> TOKEN_NEWLINE) then Asm_AddExpectedError("line break"); break
            
            
            // TOKEN_PARAM (directive)
            (TOKEN_PARAM):
                If (asmParser_isFuncActive = FALSE) then Asm_AddCodeError("Param not allowed in global scope"); break
                If (Asm_GetNextToken() <> TOKEN_IDENT) then
                    Asm_AddExpectedError("identifier")
                    break
                Endif
                
                Inc(asmParser_currFuncParamCount)
                
                If (Asm_GetNextToken() <> TOKEN_NEWLINE) then Asm_AddExpectedError("line break"); break
            
            
            // TOKEN_INSTR
            (TOKEN_INSTR): 
                If (asmParser_isFuncActive = FALSE) then Asm_AddCodeError("Instructions not allowed in global scope"); break
                Inc(asmParser_instrStreamSize)
            
            
            // TOKEN_IDENT (Line labels should be the only remaining idents)
            (TOKEN_IDENT):
                If (Asm_LexerLookAhead() <> ":") then Asm_AddCodeError("Invalid instruction"); break
                If (asmParser_isFuncActive = FALSE) then Asm_AddCodeError("Labels not allowed in global scope"); break
                
                ident = asmLexer_currLexeme
                targetIndex = asmParser_instrStreamSize - 1
                funcIndex = asmParser_currFuncIndex
                
                If (Asm_AddLabel(ident, targetIndex, funcIndex) < 0) then
                    Asm_AddCodeError("Label "+ident+" is already defined") ; break
                Endif
            
             
            // TOKEN_NEWLINE
            (TOKEN_NEWLINE): // newline - do nothing
                
            
            Default:
                Asm_AddCodeError("Undefined symbol '"+asmLexer_currLexeme+"'")
                break
        EndCase
        
        // Skip to next Line - in Pass 1 we are not concerned about any additional tokens that instrs have 
        If (Asm_LexerNextLine() = FALSE) then break    
    
    Wend
    
    


    
    // Parse Assembly - Pass 2
    // -------------------------------------------------------------------------
    
    If (asmErrors = 0) then
        
        Asm_AddMessage("Pass 2")
        
        asmParser_currInstrIndex = 0
        Asm_ResetLexer()
           
        While (TRUE)
            If (Asm_GetNextToken() = TOKEN_END_STREAM) then break
            
            Case asmLexer_currToken of    
            
                // TOKEN_FUNC (directive)
                (TOKEN_FUNC):
                    Asm_GetNextToken() // ident
                    asmParser_currFunc = Asm_GetFuncByName(asmLexer_currLexeme)
                    asmParser_isFuncActive = TRUE
                    asmParser_currFuncParamCount = 0
                    asmParser_currFuncIndex = asmParser_currFunc.index
                    
                    While (Asm_GetNextToken() = TOKEN_NEWLINE)
                    Wend
                    // Token is now open brace

                
                // TOKEN_CLOSE_BRACE
                (TOKEN_CLOSE_BRACE):
                    asmParser_isFuncActive = FALSE
                    
                    // Append a ret isntr
                    asmInstr = NewItem(instrStream)
                    asmInstr.opCode = INSTR_RET
                    asmInstr.opCount = 0
                    Inc(asmParser_currInstrIndex)
                
                
                // TOKEN_PARAM (directive)
                (TOKEN_PARAM):
                    Asm_GetNextToken()
                    ident = asmLexer_currLexeme
                    
                    // +1 for ret
                    stackIndex = -(asmParser_currFuncLocalDataSize + 2 + asmParser_currFuncParamCount + 1)
                
                    If (Asm_AddSymbol(ident, 1, stackIndex, asmParser_currFuncIndex, FALSE) < 0) then
                        Asm_AddCodeError("Variable "+ident+" is already defined") ; break
                    Endif
                    
                    Inc(asmParser_currFuncParamCount)
                    
                
                // TOKEN_INSTR
                (TOKEN_INSTR):
                
                    instruction = Asm_GetInstrByMnemonic(asmLexer_currLexeme)
                    
                    asmInstr = NewItem(instrStream)
                    asmInstr.opCode = asmInstrLookup[instruction].opCode
                    asmInstr.opCount = asmInstrLookup[instruction].opCount 
                    
                    // For each operand
                    For i = 0 to asmInstr.opCount-1
                    
                        // Lookup Operand Template
                        asmLookupOperand = IndexedItem(asmInstrLookup[instruction].opList, i)
                        
                        // Add a new Operand to instruction Steam
                        asmOperand = NewItem(asmInstr.opList) 
                        
                        // Get opType bitfield
                        opTypes = asmLookupOperand.opType
                        
                        
                        // Process Operand
                         Asm_GetNextToken()                        
                         
                         
                         
                         Case (asmLexer_currToken) of
                            
                            // TOKEN_INT
                            (TOKEN_INT):
                                If ((opTypes and OP_FLAG_INT) <> OP_FLAG_INT) then Asm_AddUnexpectedError("Integer"); break
                                asmOperand.opType = OP_TYPE_INT
                                asmOperand.iValue = ToInt(asmLexer_currLexeme)
                                

                            // TOKEN_FLOAT
                            (TOKEN_FLOAT):
                                If ((opTypes and OP_FLAG_FLOAT) <> OP_FLAG_FLOAT) then Asm_AddUnexpectedError("Float"); break
                                asmOperand.opType = OP_TYPE_FLOAT
                                asmOperand.rValue = ToReal(asmLexer_currLexeme)
                                
                            
                            // TOKEN_QUOTE
                            (TOKEN_QUOTE):
                                If ((opTypes and OP_FLAG_STRING) <> OP_FLAG_STRING) then Asm_AddUnexpectedError("String"); break
                                Asm_GetNextToken()
                                
                                Case (asmLexer_currToken) of 
                                    
                                    // Empty String - cast to integer with value 0 
                                    (TOKEN_QUOTE):
                                        asmOperand.opType = OP_TYPE_INT
                                        asmOperand.iValue = 0
                                    
                                    // Normal String
                                    (TOKEN_STRING):
                                        s = asmLexer_currLexeme
                                        sIndex = Asm_AddString(s, asmStringList)
                                        
                                        If (Asm_GetNextToken() <> TOKEN_QUOTE) then Asm_AddExpectedError('"'); break
                                        asmOperand.opType = OP_TYPE_STRING_INDEX
                                        asmOperand.iValue = sIndex
                                    
                                     
                                    
                                    Default:
                                        Asm_AddCodeError("Invalid String"); break
                                
                                EndCase
                                
                            
                            
                            // TOKEN_IDENT
                            (TOKEN_IDENT):
                            
                                Conditions
                                    
                                    // MEMREF Operand
                                    ((opTypes and OP_FLAG_MEMREF) = OP_FLAG_MEMREF):
                                    
                                        ident = asmLexer_currLexeme
                                        
                                        If Not Asm_ExistsSymbolByIdent(ident, asmParser_currFuncIndex)
                                            Asm_AddCodeError("Undefined identifier "+ident)
                                            break
                                        Endif
                                        
                                        asmSymbol = Asm_GetSymbolByIdent(ident, asmParser_currFuncIndex)
                                        sIndex = asmSymbol.stackIndex
                                        
                                        If (Asm_LexerLookAhead <> '[')
                                            // Not array
                                            
                                            // Make sure it's not defined as an array
                                            If (asmSymbol.isArray = TRUE) then
                                                Asm_AddExpectedError("array index")
                                                break
                                            Endif
                                            
                                            asmOperand.opType = OP_TYPE_ABS_STACK_INDEX
                                            asmOperand.iValue = sIndex
                                            
                                        
                                        Else
                                            // Array
                                            
                                            // Make sure it's defined as an array
                                            If (asmSymbol.isArray = FALSE)
                                                Asm_AddCodeError(asmLexer_currLexeme+" is not an array")
                                                break
                                            Endif
                                            
                                            If (Asm_GetNextToken <> TOKEN_OPEN_BRACKET) then
                                                Asm_AddExpectedError("[")
                                                break
                                            Endif
                                            
                                            
                                            // Index is either integer or variable
                                            arrayIndex = Asm_GetNextToken()
                                            
                                            Case (arrayIndex) of
                                                (TOKEN_INT):
                                                
                                                    offsetIndex = ToInt(asmLexer_currLexeme)
                                                    
                                                    If (offsetIndex < 0) or (offsetIndex >= asmSymbol.size) then
                                                        Asm_AddCodeError("Array index out of bounds")
                                                        break
                                                    Endif
                                                    
                                                    asmOperand.opType = OP_TYPE_ABS_STACK_INDEX
                                                    asmOperand.iValue = sIndex - offsetIndex // +
                                                
                                                
                                                (TOKEN_IDENT):
                                                
                                                    arrayIndexIdent = asmLexer_currLexeme
                                                    
                                                    // Ensure it's a valid variable (and not an array)
                                                    If Not Asm_ExistsSymbolByIdent(arrayIndexIdent, asmParser_currFuncIndex)
                                                        Asm_AddCodeError("Undefined identifier "+arrayIndexIdent)
                                                        break
                                                    Endif
                                                    
                                                    asmSymbol2 = Asm_GetSymbolByIdent(arrayIndexIdent, asmParser_currFuncIndex)
                                                    If asmSymbol2.isArray then
                                                        Asm_AddCodeError("Cannot use an array as an array index")
                                                        break
                                                    Endif
                                                    
                                                    offsetIndex = asmSymbol2.stackIndex
                                                    
                                                    asmOperand.opType = OP_TYPE_REL_STACK_INDEX
                                                    asmOperand.iValue = sIndex 
                                                    asmOperand.iOffsetIndex = offsetIndex 
                                                    

                                                
                                                // Non integer/variable array index
                                                Default:
                                                    Asm_AddCodeError("Invalid array index")
                                                    break
                                            EndCase
                                            
                                            // Closing bracket
                                            If (Asm_GetNextToken() <> TOKEN_CLOSE_BRACKET) then
                                                Asm_AddExpectedError("]")
                                                break
                                            Endif
                                        
                                        Endif 
                                    
                                    
                                    // LABEL Operand
                                    ((opTypes and OP_FLAG_LABEL) = OP_FLAG_LABEL):
                                        ident = asmLexer_currLexeme
                                        
                                        If (Not Asm_ExistsLabelByIdent(ident, asmParser_currFuncIndex)) then
                                            Asm_AddCodeError("Undefined line label")
                                            break
                                        Endif
                                        
                                        asmLabel = Asm_GetLabelByIdent(ident, asmParser_currFuncIndex)
                                        
                                        asmOperand.opType = OP_TYPE_INSTR_INDEX
                                        asmOperand.iValue = asmLabel.targetIndex 
                                        
                                    
                                    
                                    // FUNC Operand
                                    ((opTypes and OP_FLAG_FUNC) = OP_FLAG_FUNC):
                                        ident = asmLexer_currLexeme
                                        
                                        If (Not Asm_ExistsFuncByName(ident)) then
                                            Asm_AddCodeError("Undefined function")
                                            break
                                        Endif
                                        
                                        asmFunc = Asm_GetFuncByName(ident)
                                        
                                        asmOperand.opType = OP_TYPE_FUNC_INDEX
                                        asmOperand.iValue = asmFunc.index 
                                    
                                    
                                    // HAPI Operand
                                    ((opTypes and OP_FLAG_HAPI) = OP_FLAG_HAPI):
                                        ident = asmLexer_currLexeme
                                        
                                        asmOperand.opType = OP_TYPE_HAPI_INDEX
                                        asmOperand.iValue = Asm_AddString(ident ,asmHapiFuncList) 
                                
                                
                                    // Unkown identifier
                                    Default:
                                        Asm_AddCodeError("Unkown identifier"); break                                
                                
                                EndConditions                         
                         
                         
                            
                                
                                
                                
                         
                            default:
                                Asm_AddCodeError("Invalid operand type")
                                If (asmLexer_currToken = TOKEN_INSTR) Asm_AddMessage("("+asmLexer_currLexeme+" is a reserved instruction)")
                                If (asmLexer_currToken = TOKEN_FUNC) Asm_AddMessage("("+asmLexer_currLexeme+" is a reserved directive)")
                                If (asmLexer_currToken = TOKEN_PARAM) Asm_AddMessage("("+asmLexer_currLexeme+" is a reserved directive)")
                                If (asmLexer_currToken = TOKEN_VAR) Asm_AddMessage("("+asmLexer_currLexeme+" is a reserved directive)")
                                break                         
                         EndCase
                        
                        
                        
                        // Each operand must have a comma, unless it's the last one
                        If (i < (asmInstr.opCount-1)) then
                            If (Asm_GetNextToken() <> TOKEN_COMMA) then Asm_AddExpectedError(","); break
                        Endif
                    
                    Next
                    

                    If (asmErrors > 0) then break
                    
                    
                    // Ensure no extra operands remain
                    If (Asm_GetNextToken() <> TOKEN_NEWLINE) then Asm_AddExpectedError("line break"); break
                    Inc(asmParser_currInstrIndex)
                    
                    
            
            EndCase 
        
        Wend
    
    Endif
    
    
    // Check for presense of Main()
    isMainDefined = FALSE
    Loop asmFunc through asmFuncList
        If asmFunc.name = "MAIN" then isMainDefined = TRUE
    EndLoop
    If isMainDefined = FALSE then
        Asm_AddExpectedError("'main' function")
    Endif
    
    
    
    If (asmErrors > 0) then
        Asm_AddMessage(" ")
        Asm_AddMessage("Assembler failed with "+asmErrors+" errors")
        Asm_AddMessage("Abort.")
    Else
        
        // Add statistics
        If info then
            infoVariables = 0
            infoArrays = 0
            infoGlobals = 0
        
            Loop asmSymbol through asmSymbolList
                Inc(infoVariables)
                If asmSymbol.isArray then Inc(infoArrays)
                If asmSymbol.stackIndex >= 0 then Inc(infoGlobals) 
            EndLoop
        
            Asm_AddMessage(" ")
            Asm_AddMessage("INFO:   Lines of code processed: "+(CountList(asmSourceList)-1))
            Asm_AddMessage("INFO:    Instructions assembled: "+(CountList(instrStream)))
            Asm_AddMessage("INFO:            Host API Calls: "+(CountList(asmHapiFuncList)))
            Asm_AddMessage("INFO:                 Functions: "+(CountList(asmFuncList)))
            Asm_AddMessage("INFO:               Line Labels: "+(CountList(asmLabelList)))
            Asm_AddMessage("INFO:           String Literals: "+(CountList(asmStringList)))
            Asm_AddMessage("INFO:                 Variables: "+infoVariables+" ("+infoArrays+" arrays, "+infoGlobals+" global)")
            
            Asm_AddMessage(" ")
        Endif
        
        
        
        // ------------------------------------------------------------------------
        // Write exe
        // ------------------------------------------------------------------------

        
        // Header
        Write(result, ASM_EXE_ID)
        WriteInt(result, ASM_VERSION)
        WriteInt(result, asmParser_globalDataSize)
        
        // Stack index of _RET and _CMP
        Loop asmSymbol through asmSymbolList
            If asmSymbol.ident = "_RET" then WriteInt(result, asmSymbol.stackIndex)
        EndLoop
        Loop asmSymbol through asmSymbolList
            If asmSymbol.ident = "_CMP" then WriteInt(result, asmSymbol.stackIndex)
        EndLoop        

        
        // Sizes:
        
        // Size - String Table
        bytes = 4 // 4 for string count        
        Loop asmString through asmStringList
            bytes = bytes + 4 // string length
            bytes = bytes + Length(asmString.sString)         
        EndLoop
        WriteInt(result, bytes)
        
        
        // Size - Function Table
        bytes = 4 // 4 for function count        
        Loop asmFunc through asmFuncList
            bytes = bytes + 1 // Func name length
            bytes = bytes + Length(asmFunc.name)
            bytes = bytes + 4 // Entry
            bytes = bytes + 1 // param count
            bytes = bytes + 4 // Local data size         
        EndLoop
        WriteInt(result, bytes)


        // Size - HAPI Table
        bytes = 4 // 4 for string count        
        Loop asmString through asmHapiFuncList
            bytes = bytes + 1 // string Length (only 1 byte)
            bytes = bytes + Length(asmString.sString)         
        EndLoop
        WriteInt(result, bytes)
        
        
        // Size - Instructions
        bytes = 4 // 4 for instructions count
        Loop asmInstr through instrStream
            bytes = bytes + 1 // opCode
            bytes = bytes + 1 // opCount
            
            Loop asmOperand through asmInstr.opList
                Case asmOperand.opType of
                    (OP_TYPE_INT):              bytes = bytes + 4
                    (OP_TYPE_FLOAT):            bytes = bytes + 4
                    (OP_TYPE_STRING_INDEX):     bytes = bytes + 4
                    (OP_TYPE_INSTR_INDEX):      bytes = bytes + 4
                    (OP_TYPE_ABS_STACK_INDEX):  bytes = bytes + 4
                    (OP_TYPE_REL_STACK_INDEX):  bytes = bytes + 8
                    (OP_TYPE_FUNC_INDEX):       bytes = bytes + 4
                    (OP_TYPE_HAPI_INDEX):       bytes = bytes + 4
                EndCase
            EndLoop
        EndLoop
        WriteInt(result, bytes)

        
        
        // Write blocks
        
        // Write - String Table
        WriteInt(result, CountList(asmStringList))
        
        Loop asmString through asmStringList
            bytes = Length(asmString.sString)
            WriteInt(result, bytes)
            Write(result, asmString.sString)
            Free(asmString)         
        EndLoop
                

        // Write - Function Table
        WriteInt(result, CountList(asmFuncList))
        
        Loop asmFunc through asmFuncList
            WriteByte(result, Length(asmFunc.name))
            Write(result, asmFunc.name)             
            WriteInt(result, asmFunc.entry)
            WriteByte(result, asmFunc.paramCnt)
            WriteInt(result, asmFunc.dataSize)
            Free(asmFunc)      
        EndLoop
        

        // Write - HAPI Table
        WriteInt(result, CountList(asmHapiFuncList))
        
        Loop asmString through asmHapiFuncList
            bytes = Length(asmString.sString)
            WriteByte(result, bytes)
            Write(result, asmString.sString)
            Free(asmString)         
        EndLoop
                

        
        // Write - Instruction Stream
        WriteInt(result, CountList(instrStream))
        
        Loop asmInstr through instrStream
            WriteByte(result, asmInstr.opCode)
            WriteByte(result, asmInstr.opCount)
            
            Loop asmOperand through asmInstr.opList
                
                WriteByte(result, asmOperand.opType)
                
                Case asmOperand.opType of
                    (OP_TYPE_INT):              WriteInt(result, asmOperand.iValue)
                    (OP_TYPE_FLOAT):            WriteReal(result, asmOperand.rValue)
                    (OP_TYPE_STRING_INDEX):     WriteInt(result, asmOperand.iValue)
                    (OP_TYPE_INSTR_INDEX):      WriteInt(result, asmOperand.iValue)
                    (OP_TYPE_ABS_STACK_INDEX):  WriteInt(result, asmOperand.iValue)
                    (OP_TYPE_REL_STACK_INDEX):  WriteInt(result, asmOperand.iValue) ; WriteInt(result, asmOperand.iOffsetIndex)
                    (OP_TYPE_FUNC_INDEX):       WriteInt(result, asmOperand.iValue)
                    (OP_TYPE_HAPI_INDEX):       WriteInt(result, asmOperand.iValue)
                EndCase
            
                Free(asmOperand)
            EndLoop
            
            Free(asmInstr)
        EndLoop

        Asm_AddMessage("Assembled: "+FileSize(result)+" bytes written. Took "+(Millisecs - m)+" millisecs.")
        Asm_AddMessage("Done.")
        
    Endif
    
    
    // Free everything
    
    Loop asmFunc through asmFuncList
        Free(asmFunc)
    EndLoop
    
    Loop asmLabel through asmLabelList
        Free(asmLabel)
    EndLoop

    Loop asmString through asmStringList
        Free(asmString)
    EndLoop

    Loop asmString through asmHapiFuncList
        Free(asmString)
    EndLoop

    Loop asmSymbol through asmSymbolList
        Free(asmSymbol)
    EndLoop
    
    Loop asmInstr through instrStream
        Free(asmInstr)
    EndLoop
    
    Loop asmSourceLine through asmSourceList
        Free(asmSourceLine)
    EndLoop
    
End

Function Asm_AddString(sString: String, sList: List of asmStrings) : Integer
Var
    asmString: ^asmStrings
Begin
    
    Loop asmString through sList
        If asmString.sString = sString then
            result = ItemInList(sList, asmString)
            exit
        Endif
    EndLoop
    
    asmString = NewItem(sList)
    asmString.sString = sString
    result = ItemInList(sList, asmString) 
     
End

Function Asm_AddFunc(name: String, entry: Integer) : Integer
{
    Asm_AddFunc
    Returns -1 on error, or the function's index on success
}
Var
    asmFunc : ^asmFuncs
Begin
    asmFunc = Asm_GetFuncByName(name)
    If (ItemInList(asmFuncList, asmFunc) > -1) then
        result = -1
        exit
    Endif
    
    asmFunc = NewItem(asmFuncList)
    asmFunc.name = name
    asmFunc.entry = entry
    asmFunc.index = CountList(asmFuncList) - 1
    result = asmFunc.index
    
End

Function Asm_GetFuncByName(name: String) : ^asmFuncs
{
    Asm_GetFuncByName
    Returns a function list item with a matching name  
    Use ItemInList(asmFuncList, asmFunc) to check for valid result
}
Var
    asmFunc : ^asmFuncs
Begin
    Loop asmFunc through asmFuncList
        If asmFunc.name = name then
            result = asmFunc
            Exit
        Endif
    EndLoop
End

Function Asm_ExistsFuncByName(name: String) : Boolean
Var
    asmFunc : ^asmFuncs
Begin
    result = FALSE

    Loop asmFunc through asmFuncList
        If asmFunc.name = name then
            result = TRUE
            Exit
        Endif
    EndLoop
End

Function Asm_AddSymbol(ident: String, size: Integer, stackIndex: Integer, funcIndex: Integer, isArray: Boolean) : Integer
{
    Asm_AddSymbol
    Returns -1 on error, or the symbol's index on success
}
Var
    asmSymbol : ^asmSymbols
Begin
    asmSymbol = Asm_GetSymbolByIdent(ident, funcIndex)
    If (ItemInList(asmSymbolList, asmSymbol) > -1) then
        result = -1
        exit
    Endif
    
    asmSymbol = NewItem(asmSymbolList)
    asmSymbol.ident      = ident
    asmSymbol.size       = size
    asmSymbol.stackIndex = stackIndex
    asmSymbol.funcIndex  = funcIndex
    asmSymbol.isArray    = isArray 
    asmsymbol.index = CountList(asmSymbolList) - 1
    result = asmSymbol.index
End

Function Asm_GetSymbolByIdent(ident: String, funcIndex: Integer) : ^asmSymbols
{
    Asm_GetSymbolByIdent
    Returns a symbol list item with a matching name in a valid scope  
    Use ItemInList(asmSymbolList, asmSymbol) to check for valid result
}
Var
    asmSymbol : ^asmSymbols
Begin
    Loop asmSymbol through asmSymbolList
        If (asmSymbol.funcIndex = funcIndex) or (asmSymbol.stackIndex >= 0) then // scope
            If asmSymbol.ident = ident then
                result = asmSymbol
                Exit
            Endif
        Endif
    EndLoop
End

Function Asm_ExistsSymbolByIdent(ident: String, funcIndex: Integer) : Boolean
Var
    asmSymbol : ^asmSymbols
Begin
    result = FALSE

    Loop asmSymbol through asmSymbolList
        If (asmSymbol.funcIndex = funcIndex) or (asmSymbol.stackIndex >= 0) then // scope
            If asmSymbol.ident = ident then
                result = TRUE
                Exit
            Endif
        Endif
    EndLoop
End

Function Asm_AddLabel(ident: String, targetIndex: Integer, funcIndex: Integer) : Integer
{
    Asm_AddLabel
    Returns -1 on error, or the label's index on success
}
Var
    asmLabel : ^asmLabels
Begin
    asmLabel = Asm_GetLabelByIdent(ident, funcIndex)
    If (ItemInList(asmLabelList, asmLabel) > -1) then
        result = -1
        exit
    Endif
    
    asmLabel = NewItem(asmLabelList)
    asmLabel.ident = ident
    asmLabel.targetIndex = targetIndex
    asmLabel.funcIndex = funcIndex
    asmLabel.index = CountList(asmLabelList) - 1
    result = asmLabel.index
    
End

Function Asm_GetLabelByIdent(ident: String, funcIndex: Integer) : ^asmLabels
{
    Asm_GetLabelByIdent
    Returns a label list item with a matching name in a valid scope  
    Use ItemInList(asmLabelList, asmLabel) to check for valid result
}
Var
    asmLabel : ^asmLabels
Begin
    Loop asmLabel through asmLabelList
        If (asmLabel.funcIndex = funcIndex) then // scope
            If asmLabel.ident = ident then
                result = asmLabel
                Exit
            Endif
        Endif
    EndLoop
End

Function Asm_ExistsLabelByIdent(ident: String, funcIndex: Integer) : Boolean
Var
    asmLabel : ^asmLabels
Begin
    result = FALSE

    Loop asmLabel through asmLabelList
        If (asmLabel.funcIndex = funcIndex) then // scope
            If asmLabel.ident = ident then
                result = TRUE
                Exit
            Endif
        Endif
    EndLoop
End

Function Asm_AddInstrLookup(mnemonic: String, opCode: Integer, opCount: Integer): Integer
{
    Asm_AddInstrLookup
    Returns the index of a new Instr lookup
    Returns -1 if too many items have been added.
    Returns -2 if you are trying to overwrite an existing mnemonic
}
Var
    asmOperand : ^asmOperands
    index: Integer
    i: Integer
Begin
    index = opCode

    If index >= MAX_INSTRS then
        result = -1
        MessageBox("Instruction Lookup Table full - cannot add additional instructions.", "ASM")
        exit
    Endif
    
    If (asmInstrLookup[index].mnemonic <> "") then
        result = -2
        MessageBox("Error adding to Instruction Lookup Table - mnemonic '"+mnemonic+"' tried to overwrite existing mnemonic '"+asmInstrLookup[index].mnemonic+"'.", "ASM")
        exit
    Endif 
    
    asmInstrLookup[index].mnemonic = mnemonic
    asmInstrLookup[index].opCode   = opCode
    asmInstrLookup[index].opCount  = opCount
    
    For i = 0 to opCount-1
        asmOperand = NewItem(asmInstrLookup[index].opList)
    Next    
    
    result = index
End

Procedure Asm_SetInstrLookupOpType(instrIndex: Integer, opIndex: Integer, types: Integer)
{
    Asm_SetInstrLookupOpType
}
Var
    asmOperand : ^asmOperands
Begin
    If instrIndex >= 0 then 
        asmOperand = IndexedItem(asmInstrLookup[instrIndex].opList, opIndex)
        asmOperand.opType = types
    Endif
End 

Function Asm_GetInstrByMnemonic(mnemonic: String) : Integer
{
    Asm_GetInstrByMnemonic
    Returns -1 for not found, otherwise the index of the item in the asmInstrLookup array
}
Var
    i: Integer
Begin
    For i = 0 to (MAX_INSTRS-1)
        If asmInstrLookup[i].mnemonic = mnemonic then
            result = i
            exit
        Endif
    Next
    
    result = -1
End

Function Asm_GetNextToken() : Integer
Var
    l, i: Integer
    c: String
    continue: Boolean = FALSE
Begin

    asmLexer_index0 = asmLexer_index1
    result = TOKEN_INVALID
    
    // Advance to next Line as necessary
    l = Length(asmCurrLine.s)
    If asmLexer_index0 >= l then
        If Asm_LexerNextLine() = FALSE then
            result = TOKEN_END_STREAM
            exit
        Endif
    Endif
    
    If (asmLexer_currState = LEXER_STATE_END_STR) then asmLexer_currState = LEXER_STATE_NO_STR

    // Scan through all potential whitespace    
    If (asmLexer_currState <> LEXER_STATE_IN_STR) then
    
        While (TRUE)
            c = Mid(asmCurrLine.s, asmLexer_index0+1, 1)
            If Asm_CharIsWhiteSpace(c) = FALSE then break
            Inc(asmLexer_index0)
        Wend
        
    Endif
    
    asmLexer_index1 = asmLexer_index0
    
    
    
    // Scan through lexeme until a delimiter is met, incrementing index1    
    While (TRUE)
        If (asmLexer_currState = LEXER_STATE_IN_STR) then
            continue = FALSE
        
            l = Length(asmCurrLine.s)
        
            // Strings cannot span multiple lines
            If asmLexer_index1 >= l then
                asmLexer_currToken = TOKEN_INVALID
                result = asmLexer_currToken
                exit 
            Endif
            
            c = Mid(asmCurrLine.s, asmLexer_index1+1, 1)
            
            // Backslash
            If c = '\' then
                asmLexer_index1 = asmLexer_index1 + 2
                continue = TRUE
            Endif  
            
            // Double quote
            If (continue = FALSE) and (c = '"') then
                break            
            Endif
            
            If (continue = FALSE) then Inc(asmLexer_index1) 
        
        
        Else // Not a string
            
            l = Length(asmCurrLine.s)
            If asmLexer_index1 >= l then break
            
            c = Mid(asmCurrLine.s, asmLexer_index1+1, 1)
            
            If Asm_CharIsDelim(c) then break
            
            Inc(asmLexer_index1)
        
        Endif        
    Wend
    
    If (asmLexer_index1 - asmLexer_index0) = 0 then Inc(asmLexer_index1)
    asmLexer_currLexeme = ""
    asmLexer_currToken = TOKEN_INVALID
    
    
    
    // Lexeme Isolatted, lies between index0 and index1
    For i = asmLexer_index0 to asmLexer_index1-1
    
        If (asmLexer_currState = LEXER_STATE_IN_STR) then
            If Mid(asmCurrLine.s, asmLexer_index1+1, 1) = '\' then
                Inc(i)
            Endif
        Endif
        
        asmLexer_currLexeme = asmLexer_currLexeme + Mid(asmCurrLine.s, i+1, 1)
    
    Next
    
    If (asmLexer_currState <> LEXER_STATE_IN_STR) then asmLexer_currLexeme = Upper(asmLexer_currLexeme)
    l = Length(asmLexer_currLexeme) 
    
        
    // Token Identification
    
    // A string token
    If (l > 1) or (Left(asmLexer_currLexeme, 1) <> '"') then
        If (asmLexer_currState = LEXER_STATE_IN_STR) then
            asmLexer_currToken = TOKEN_STRING
            result = asmLexer_currToken
            exit
        Endif
    Endif
    
    If l = 1 then
        // Single-character tokens
        
        Conditions
            
            (asmLexer_currLexeme = '"'):
                Case asmLexer_currState of
                    (LEXER_STATE_NO_STR):  asmLexer_currState = LEXER_STATE_IN_STR
                    (LEXER_STATE_IN_STR):  asmLexer_currState = LEXER_STATE_END_STR
                EndCase
                asmLexer_currToken = TOKEN_QUOTE
            
            (asmLexer_currLexeme = ','): asmLexer_currToken = TOKEN_COMMA
            (asmLexer_currLexeme = ':'): asmLexer_currToken = TOKEN_COLON
            (asmLexer_currLexeme = '['): asmLexer_currToken = TOKEN_OPEN_BRACKET
            (asmLexer_currLexeme = ']'): asmLexer_currToken = TOKEN_CLOSE_BRACKET
            (asmLexer_currLexeme = '{'): asmLexer_currToken = TOKEN_OPEN_BRACE
            (asmLexer_currLexeme = '}'): asmLexer_currToken = TOKEN_CLOSE_BRACE
            (asmLexer_currLexeme = Chr(10)): asmLexer_currToken = TOKEN_NEWLINE
            
            (asm_CharIsNumeric(asmLexer_currLexeme)):   asmLexer_CurrToken = TOKEN_INT
            (asm_CharIsIdent(asmLexer_currLexeme)): asmLexer_CurrToken = TOKEN_IDENT
            
        EndConditions
        
        
    Else
        // Multi-character tokens
        
        Conditions
            (asmLexer_currLexeme = "VAR"):   asmLexer_CurrToken = TOKEN_VAR
            (asmLexer_currLexeme = "FUNC"):  asmLexer_CurrToken = TOKEN_FUNC
            (asmLexer_currLexeme = "PARAM"): asmLexer_CurrToken = TOKEN_PARAM
            
            (Asm_GetInstrByMnemonic(asmLexer_currLexeme) >= 0): asmLexer_CurrToken = TOKEN_INSTR
            
            (asm_StringIsInt(asmLexer_currLexeme)):   asmLexer_CurrToken = TOKEN_INT
            (asm_StringIsFloat(asmLexer_currLexeme)): asmLexer_CurrToken = TOKEN_FLOAT
            (asm_StringIsIdent(asmLexer_currLexeme)): asmLexer_CurrToken = TOKEN_IDENT
        EndConditions
        
    Endif
    
    result = asmLexer_currToken

End

Function Asm_LexerNextLine() : Boolean
Begin

    If asmCurrLine = LastItem(asmSourceList) then result = FALSE ; exit
    
    asmCurrLine = NextItem(asmSourceList, asmCurrLine)
    
    asmLexer_currState = LEXER_STATE_NO_STR
    asmLexer_index0 = 0
    asmLexer_index1 = 0
    
    result = TRUE

End

  Procedure Asm_ResetLexer()
Begin
    asmCurrLine = FirstItem(asmSourceList)    
    asmLexer_index0 = 0
    asmLexer_index1 = 0
    asmLexer_currToken = TOKEN_INVALID
    asmLexer_currState = LEXER_STATE_NO_STR

End

Function Asm_LexerLookAhead() : String
Var
    index: Integer
    c: String = ""
    tmp_asmCurrLine : ^asmSourceLines
Begin
    result = ""

    index = asmLexer_index1
    tmp_asmCurrLine = asmCurrLine
    
    
    If (asmLexer_currState <> LEXER_STATE_IN_STR)
        While (TRUE)
        
            If (index >= Length(asmCurrLine.s)) then
                If asmCurrLine = LastItem(asmSourceList) then
                    asmCurrLine = tmp_asmCurrLine
                    exit
                Else
                    asmCurrLine = NextItem(asmSourceList, asmCurrLine)
                Endif
                
                index = 0
            Endif
            
            c = Mid(asmCurrLine.s, index+1, 1)
            If (Asm_CharIsWhiteSpace(c) = FALSE) then break
            Inc(index)
                    
        Wend
    Endif
    
    result = c    

    // Reset line
    asmCurrLine = tmp_asmCurrLine
    
End






Procedure Asm_Init() ; export
{
    Asm_Init
    Initialised the assembler:
        * Adds Instruction Lookups
}
Var
    i: Integer
Begin
    If asmDoneInit = TRUE then exit
    asmDoneInit = TRUE
    
    Asm_AddMessage("Initializing Assembler...")
    
    For i = 0 to (MAX_INSTRS-1)
        New(asmInstrLookup[i])
    Next
    
    i = Asm_AddInstrLookup("MOV", INSTR_MOV, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)
    
    i = Asm_AddInstrLookup("PUSH", INSTR_PUSH, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)
    
    i = Asm_AddInstrLookup("POP", INSTR_POP, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    
    i = Asm_AddInstrLookup("ADD", INSTR_ADD, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF) 

    i = Asm_AddInstrLookup("SUB", INSTR_SUB, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)
    
    i = Asm_AddInstrLookup("MUL", INSTR_MUL, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)
    
    i = Asm_AddInstrLookup("DIV", INSTR_DIV, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("MOD", INSTR_MOD, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("ABS", INSTR_ABS, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("EXP", INSTR_EXP, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("NEG", INSTR_NEG, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    
    i = Asm_AddInstrLookup("INC", INSTR_INC, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("DEC", INSTR_DEC, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    
    i = Asm_AddInstrLookup("AND", INSTR_AND, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("OR", INSTR_OR, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("XOR", INSTR_XOR, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("NOT", INSTR_NOT, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("SHL", INSTR_SHL, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("SHR", INSTR_SHR, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("CALL", INSTR_CALL, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_FUNC)

    i = Asm_AddInstrLookup("RET", INSTR_RET, 0)
    
    i = Asm_AddInstrLookup("HAPI", INSTR_HAPI, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_HAPI)
    
    i = Asm_AddInstrLookup("JMP", INSTR_JMP, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_LABEL)

    i = Asm_AddInstrLookup("CMP", INSTR_CMP, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)

    i = Asm_AddInstrLookup("JE", INSTR_JE, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_LABEL)

    i = Asm_AddInstrLookup("JNE", INSTR_JNE, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_LABEL)

    i = Asm_AddInstrLookup("JG", INSTR_JG, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_LABEL)

    i = Asm_AddInstrLookup("JL", INSTR_JL, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_LABEL)

    i = Asm_AddInstrLookup("JGE", INSTR_JGE, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_LABEL)

    i = Asm_AddInstrLookup("JLE", INSTR_JLE, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_LABEL)

    i = Asm_AddInstrLookup("CONCAT", INSTR_CONCAT, 2)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    Asm_SetInstrLookupOpType(i, 1, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)   

    i = Asm_AddInstrLookup("TERM", INSTR_TERM, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_INT OR OP_FLAG_FLOAT OR OP_FLAG_STRING OR OP_FLAG_MEMREF)
    
    i = Asm_AddInstrLookup("FLOAT", INSTR_FLOAT, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    
    i = Asm_AddInstrLookup("INT", INSTR_INT, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)
    
    i = Asm_AddInstrLookup("STRING", INSTR_STRING, 1)
    Asm_SetInstrLookupOpType(i, 0, OP_FLAG_MEMREF)

End

Function Asm_StripComments(s: String) : String
Var
    bInStr: Boolean = FALSE
    i: Integer
    c, lastc: String = ""
Begin
    If (Instr(s, ';') = 0) then
        result = s
        exit
    Endif

    For i = 1 to Length(s)
        lastc = c
        c = Mid(s, i, 1)
        
        If (c = '"') and (lastc <> '\') then bInStr = Not bInStr
        
        If (c = ';') then
            If bInStr = FALSE then
                result = Left(s, i-1)
                exit
            Endif
        Endif 
    Next
    
    result = s

End

Function Asm_StripWhitespace(s: String) : String
Var
    i, l: Integer
    c, r: String
    
    lw, rw: Integer = 1
    ld: Boolean = FALSE
Begin
    result = s
    l = Length(s)
    If l = 0 then exit
    
    For i = 1 to l    
        c = Mid(s, i, 1)
        
        If Asm_CharIsWhitespace(c) then
            If ld = FALSE then
                lw = i + 1
            Endif
        Else
            ld = TRUE
            rw = i
        Endif
        
    Next
    
    result = Mid(s, lw, rw-lw+1)
    
    
End

Function Asm_CharIsNumeric(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c)    
    If (a >= 48) and (a <= 57) then result = TRUE else result = FALSE    
End

Function Asm_CharIsIdent(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c)
    result = FALSE    
    If (a >= 48) and (a <= 57) then result = TRUE  ; exit // 0 - 9
    If (a >= 65) and (a <= 90) then result = TRUE  ; exit // A - Z
    If (a >= 97) and (a <= 122) then result = TRUE ; exit // a - z
    If (a = 95) then result = TRUE // _    
End

Function Asm_CharIsDelim(c: String) : Boolean
Begin
    result = FALSE    
    If (c = ':') then result = TRUE  ; exit
    If (c = ',') then result = TRUE  ; exit
    If (c = '"') then result = TRUE  ; exit
    If (c = '[') then result = TRUE  ; exit
    If (c = ']') then result = TRUE  ; exit
    If (c = '{') then result = TRUE  ; exit
    If (c = '}') then result = TRUE  ; exit
    If (Asc(c) = 10) then result = TRUE  ; exit
    result = Asm_CharIsWhitespace(c)  
End

Function Asm_CharIsWhitespace(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c) // Tab = 9, Space = 32    
    If (a = 9) or (a = 32) then result = TRUE else result = FALSE
End    

Function Asm_StringIsInt(s: String) : Boolean ; export
Var
    l, i: Integer
    c: String
Begin
    result = FALSE
    l = Length(s)
    If l < 1 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (Asm_CharIsNumeric(c) or (c = '-')) then exit
        If (i <> 1) and (c = '-') then exit 
    Next
    
    result = TRUE        
End

Function Asm_StringIsFloat(s: String) : Boolean ; export
Var
    l, i: Integer
    radix: Integer = 0
    c: String
Begin
    result = FALSE
    l = Length(s)
    If l < 1 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (Asm_CharIsNumeric(c) or (c = '-') or (c = '.')) then exit
        If (i <> 1) and (c = '-') then exit
        
        If (c = '.') then
            Inc(radix)
            If i = l then exit
        Endif
        If radix > 1 then exit 
    Next
    
    result = TRUE        
End

Function Asm_StringIsIdent(s: String) : Boolean
Var
    l, i: Integer
    c: String
Begin
    result = FALSE
    l = Length(s)
    If l < 1 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If (i = 1) and Asm_CharIsNumeric(c) then exit
        If Asm_CharIsIdent(c) = FALSE then exit 
    Next
    
    result = TRUE        
End

Procedure Asm_AddExpectedError(s: String)
Begin
    Asm_AddCodeError("Expected "+s)
End

Procedure Asm_AddUnexpectedError(s: String)
Begin
    Asm_AddCodeError("Unexpected "+s)
End

Procedure Asm_AddCodeError(s: String)
Begin
    Inc(asmErrors)

    Asm_AddMessage("Error: "+s+" at line "+(ItemInList(asmSourceList, asmCurrLine)-1))

    s = Replace(asmCurrLine.s, Chr(9), ' ', FALSE) // tabs -> spaces
    s = Replace(s, Chr(10), '', FALSE) // newline
    
    Asm_AddMessage(s)
    Asm_AddMessage(RepStr(' ', asmLexer_index0)+'^')
End

Procedure Asm_AddMessage(s: String)
Var
    asmResultString: ^asmResultStrings
Begin
    asmResultString = NewItem(asmResultStringList)
    If Assigned(asmResultString) then
        asmResultString.s = s
    Else
        MessageBox("There was an error adding an error.")
    Endif
End

Procedure Asm_ClearMessages() ; export
Var
    asmResultString: ^asmResultStrings
Begin
    Loop asmResultString through asmResultStringList
        Free(asmResultString)
    EndLoop
    asmErrors = 0
End


Begin
End
